/*
 * $Id: vegpopd.c,v 1.3 1998/10/18 11:48:45 root Exp root $
 * Veghead 1998 
 * Inspired by the sharp minds at Direct Connection
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <errno.h>
#include <time.h>
#include "vegpopd.h"

int debug=0;
int sock;
pgroup_t *TESTP;
	
main(int argc,char**argv)
{
	struct protoent *pe;
	struct sockaddr_in addr,raddr;
	int got2,got,fd,res,portnum=PORTNUM,
	    daemon=1;
	char ib[MAXMSGSIZE*2];
	pgroup_t *group_top=NULL;
	client_t *client_top=NULL;	
	
	argv++;
	while (argc>1) {
	   switch(**argv) {
	       case('-'):
	     switch((*argv)[1]) {
	         case('i'):daemon=0;break;
	         case('d'):debug++;break;
	         case('p'):if (argc>2) {argv++;argc--;portnum=atoi(*argv);}break;
	           default:usage();
	         }
	      break;
	       default:usage();
	       }
	  argv++;
	  argc--;
	  }
	
	if (debug) printf("Local port:%d\n",portnum);
	
	init_conf(&group_top,&client_top);
	if (debug) printf("Allocated %s\n",group_top->grname);
	
	if (daemon) {
	   if ((fd=open("/dev/tty",O_RDONLY|O_NOCTTY))!=-1) {
	      ioctl(fd,TIOCNOTTY);
	      }
	   if (fork()!=0) {
	      exit(0);
	      }
	}
	pe=getprotobyname("UDP");
	if ((sock=socket(AF_INET,SOCK_DGRAM,pe->p_proto))==-1) {
	   leave("unable to create socket");
	   }
	addr.sin_family=AF_INET;
	addr.sin_port=htons(portnum);
	addr.sin_addr.s_addr=htons(INADDR_ANY);
	if (res=bind(sock,(struct sockaddr *)&addr,sizeof(addr))==-1) {
	   leave("bind");
	  }
	got2=sizeof(raddr);
	while (1) {
		if ((got=recvfrom(sock,ib,1024,0,(struct sockaddr *)&raddr,
								&got2))!=-1) {
			printf("%s\n",inet_ntoa(raddr.sin_addr.s_addr));
			if (memcmp(ib,VEGPOPID,strlen(VEGPOPID))==0) {
				switch (ib[3]) {
					case ('I') :break;
					case ('M') :
						proc_msg(&group_top,&client_top,
							ib,&raddr);
							break;
					default    :if (debug) puts("Fish");
							break;
				}
			}
		}
	}
}
	

/*
 * proc_msg
 * Decides what to do with an incoming message
 */
proc_msg(pgroup_t **topg,client_t **topc,char *packet,struct sockaddr_in *sad)
{
vpmsg_t *pvm;
client_t *cl;
pgroup_t *pg;
int res;

	pvm=(vpmsg_t *)packet;
	stringify(pvm->fromnick,sizeof(pvm->fromnick));
	stringify(pvm->tonick,sizeof(pvm->tonick));
	stringify(pvm->data,sizeof(pvm->data));
	if (debug) printf("Packet1>%s:%s:%s:%s\n",
		inet_ntoa(sad->sin_addr.s_addr), 
		pvm->fromnick,pvm->tonick,pvm->data);
	if ((cl=find_client(*topc,pvm->fromnick))==NULL) return(0);
	pvm->type='R';
	res=sendto(sock,packet,SIZE_VPMSGHDR,0,(struct sockaddr *)sad,
						sizeof(struct sockaddr_in));
	if (res<=0) perror("sendto");
	if ((cl=find_client(*topc,pvm->tonick))!=NULL) {
		send_vpmsg_client(sock,cl,pvm,SIZE_VPMSGHDR+strlen(pvm->data));
	} else {
		if ((pg=find_group(*topg,pvm->tonick))!=NULL) {
			send_vpmsg_group(*topc,sock,pg,pvm,
					SIZE_VPMSGHDR+strlen(pvm->data));
		} else {
			if (debug) {
				printf("No such entity %s\n",pvm->tonick);
			}
		}
	}
	
}


/*
 * send_vpmsg_group
 * walks a group and sends a message to all members using
 * send_vpmsg_client
 */
send_vpmsg_group(client_t *topc,int sock, pgroup_t *pg,vpmsg_t vp,int msglen)
{
gmem_t *gm;
client_t *cl;

	gm=pg->active;
	while (gm!=NULL) {
		if((cl=find_client(topc,gm->clientname))!=NULL) {
			send_vpmsg_client(sock,cl,vp,msglen);
		}
		gm=gm->next_gm;
	}
}



/*
 * send_vpmsg_client
 * Can't you guess what this does ?
 */
send_vpmsg_client(int sock,client_t *cl,vpmsg_t *vp,int msglen)
{
int res;

	vp->type='M';
	if (debug) printf("Sending to %s at %s:%d\n",
		cl->clientname,inet_ntoa(cl->cliaddr.sin_addr.s_addr),
		ntohs(cl->cliaddr.sin_port));
	res=sendto(sock,vp,msglen,0,(struct sockaddr *)&(cl->cliaddr),
                                                sizeof(struct sockaddr_in));
        if (res<=0) perror("sendto");	
	return(1);
}



/*
 * addtogroup
 * adds a client to a group.
 * Checks that the client and the group exist, mallocs a gmem_t,
 * walks the group and shoves the new client in at the end.
 */
addtogroup(pgroup_t **topg,client_t **topc,char *cliname, char *grname)
{
pgroup_t *pg;
client_t *cl;
gmem_t *gm,*newgm;
int num_members=0;

	if ((cl=find_client(*topc,cliname))==NULL) return(0);
	if ((pg=find_group(*topg,grname))==NULL) return(0);
	if ((newgm=(gmem_t *)malloc(sizeof(gmem_t)))==NULL) return(0);
	strncpy(newgm->clientname,cliname,MAXNAMELEN);
	if (pg->active==NULL) {
		pg->active=newgm;
	} else {
		gm=pg->active;
		while ((gm->next_gm!=NULL)&&(num_members<MAXGROUPSIZE)) {
			num_members++;
			gm=gm->next_gm;
		}
		gm->next_gm=newgm;
	}
return(1);
}


/* 
 * init_conf
 * Really dodgy routine for reading the cf file and adding groups and
 * members. It will eventually be replaced by something that works properly.
 * One day I'll splash out and by a book on yacc.
 */
void
init_conf(pgroup_t **topg,client_t **topc)
{
struct sockaddr_in fred;
FILE *cfp;
char *c;
int user=0,group=0;
char arg1[CMDLEN],arg2[CMDLEN];

	if ((cfp=fopen(TALKCF,"r"))==NULL) {
		perror("fopen");
		exit(1);
	}
	while (!feof(cfp)) {
		freadstring(cfp,arg1,CMDLEN);
		if (strcmp(arg1,"GROUP:")==0) {
			freadgroup(cfp,topg,topc);
		} else {
			freadstring(cfp,arg2,CMDLEN);
			if (*arg1) parse_add_client(arg1,arg2,topc);
		}
	}
	//exit(0);
}
	

/*
 * freadgroup
 * reads in group members, separated by whitespace, and terminated with a
 * single '.'
 */
freadgroup(FILE *fp,pgroup_t **topg,client_t **topc)
{
int groupin=1;
pgroup_t *gp;
char *tp;
char grname[MAXNAMELEN];
char memname[MAXNAMELEN];
char errorst[512];

	freadstring(fp,grname,MAXNAMELEN);
	if ((gp=find_group(*topg,grname))==NULL) 
				gp=new_pgroup(topg,grname);
	while ((!feof(fp))&&groupin) {
		freadstring(fp,memname,MAXNAMELEN);
		if (debug) printf("%s ",memname);
		if (tp=(char *)strchr(memname,'.')) {
			*tp='\0';
			groupin=0;
			if (debug) puts("Group ends");
		} else {
			if (find_client(*topc,memname)==0)  {
				strncpy(errorst,memname,sizeof(errorst));
				strncat(errorst,":No Such Client",
					sizeof(errorst)-strlen(errorst));
				bomb(errorst);
			}
			if (!addtogroup(topg,topc,memname,grname)) {
				strncpy(errorst,memname,sizeof(errorst));
                                strncat(errorst,":Can't add to group",
                                        sizeof(errorst)-strlen(errorst));
                                bomb(errorst);
			}

		}
	}
	
	
}



/*
 * freadstring
 * skips leading whitespace, reads a string into buf
 */
freadstring(FILE *fp,char *buf,int maxlen) 
{
char c;
	while((!feof(fp))&&(isspace(c=(char)fgetc(fp)))); 
	while((!feof(fp))&&(!isspace(c))) {
		if (maxlen>1) {
			*buf++=toupper(c);
			maxlen--;
		}
		c=(char)fgetc(fp);
	}
*buf='\0';
}	


/*
 * new_pgroup
 * Creates a new pgroup, and attaches to the groups list
 */
pgroup_t *
new_pgroup(pgroup_t **topg,char *newname)
{
pgroup_t *pg,*tp;
	pg=(pgroup_t *)malloc(sizeof(pgroup_t));
	if (pg==NULL) {
		return((pgroup_t *)NULL);
	}
	strncpy(pg->grname,newname,MAXNAMELEN);
	pg->bar=NULL;
	pg->active=NULL;
	pg->next_pg=NULL;
	tp=*topg;
	if (tp==NULL) {
		*topg=pg;
		if (debug) puts("New top level group");
	} else {
		while(tp->next_pg!=NULL) {
			tp=tp->next_pg;
		}
		tp->next_pg=pg;
		if (debug) printf("Added group %s\n",newname);
	}
	return(pg);
}



parse_add_client(char *name,char *adr,client_t **topc)
{
int i,port=PORTNUM;
u_long ula;
struct sockaddr_in so;
char *pp;
	if ((strlen(name)>MAXNAMELEN-1)||(!(*name))) 
			bomb("Client name invalid");
	for (i=0;i<strlen(name);i++) if (!isalnum(name[i])) 
			bomb("Client name contains invalid characters");
	if ((pp=(char *)strchr(adr,':'))!=NULL) {
		*pp='\0';
		if (*(++pp)) port=atoi(pp);
	}
	if ((ula=inet_addr(adr))==-1) 
			bomb("Invalid address");
	so.sin_addr.s_addr=ula;
	so.sin_port=htons(port);
	so.sin_family=AF_INET;
	return(new_client(topc,name,&so));
}



bomb(char *bm)
{
fprintf(stderr,"%s\n",bm);
exit(1);
}


/*
 * find_client
 * walks the client list to find a specific cliname, and then
 * returns a pointer to it. Ornf failure returns NULL.
 */
client_t *
find_client(client_t *cl,char *cliname)
{
client_t *clp;
	if (cl==NULL) return((client_t *)NULL);
	clp=cl;
	while (clp!=NULL) {
		if (strcmp(clp->clientname,cliname)==0) return(clp);
		clp=clp->next_c;
	}	
return((client_t *)NULL);
}


/*
 * client_exists
 * does the named client exist ?
 * this is now just runs a find_client.
 */
client_exists(client_t *cl, char *cliname)
{
	if (find_client(cl,cliname)!=NULL) return(1); else return(0);
}


pgroup_t *
find_group(pgroup_t *gp,char *grname)
{
pgroup_t *gpt;
	if (gp==NULL) return((pgroup_t*)NULL);
	gpt=gp;
	while (gpt!=NULL) {
		if (strcmp(gpt->grname,grname)==0) return(gpt);
		gpt=gpt->next_pg;
	}
return((pgroup_t *)NULL);
}


/* 
 * new_client
 * creates a new client, and tacks it onto the end of the client
 * list.
 */
new_client(client_t **cl,char *cliname, struct sockaddr_in *adr)
{
	int clicount=0;
	client_t **cp;
	if (client_exists(*cl,cliname)) {
		if (debug) printf("%s exists already\n",cliname);
		return(0);
	}
	cp=cl;
	while ((clicount<=MAXNUMCLIENTS)&&(*cp!=NULL)) {
		clicount++;
		cp=&((*cp)->next_c);	
	}
	if (clicount>MAXNUMCLIENTS) return(0);
	*cp=(client_t *)malloc(sizeof(client_t));
	if (*cp==NULL) return(0);
	(*cp)->next_c=NULL;
	(*cp)->status&=1;
	(*cp)->t=time(NULL);
	memcpy(&((*cp)->cliaddr),adr,sizeof(struct sockaddr_in));
	memcpy(&((*cp)->clientname),cliname,MAXNAMELEN);
	if (debug) printf("%s client %d %d\n",cliname,clicount,(*cp)->t);
	return(1);
}
	
	
	
void
leave(char *out)
{
	perror(out);
	exit(1);
}
	

/*
pgroup_t *
find_group(pgroup_t *top,char *gname)
{
pgroup_t *pg;
pg=top;
while (pg!=NULL) {
	if (strcmp(pg->grname,gname)==0) return(pg);
	pg=pg->next_pg;
	}
return(NULL);
}
*/

	
/*From winpub	
addclient(struct sockaddr_in *ad, struct client *clidb)
{
	int i=0,ok=0,added=-1;
	time_t timenow;
	timenow=time(NULL);
	for (i=0;i<CLISIZE;i++) {
	   if ((debug)&&(clidb[i].status=='1')) {
	      printf("%s %d\n",inet_ntoa(clidb[i].cliaddr.sin_addr),i);
	   }
	   if (clidb[i].status!='1') {
	     added=i;
	   } 
	   else if ((timenow-clidb[i].t) >STALETIME) {
	      clidb[i].status='0';
	      added=i;
	      if (debug) printf("Timeout %d %d %d\n",i,timenow,clidb[i].t);
	      }
	   else if (memcmp(&clidb[i].cliaddr.sin_addr,&ad->sin_addr,sizeof(clidb[i].cliaddr.sin_addr))==0) {
	      ok=1;
	      clidb[i].t=timenow;
	      clidb[i].cliaddr.sin_port=ad->sin_port;
	      added=i;
	      if (debug) printf("Already found at %d - %d - Port:%d\n",i,
	                        sizeof(clidb[i].cliaddr),clidb[i].cliaddr.sin_port);
	      } 
	   }
	if ((!ok)&&(added>-1)) {
	   clidb[added].status='1';
	   clidb[added].t=timenow;
	   memcpy(&clidb[added].cliaddr,ad,sizeof(clidb[added].cliaddr));
	   if (debug) printf("Added %d - %d\n",added,sizeof(clidb[added].cliaddr));
	   }
	if (debug) printf("%s is here! %d \n",inet_ntoa(clidb[added].cliaddr.sin_addr),timenow);
}
*/	
	
	
	
pubsend(int sock,char *msg,int size,pgroup_t *pg)
{
	int i,res;
	time_t timenow;
	client_t *cp;

	timenow=time(NULL);
	if (debug) puts("Pubsend");	
	msg[0]='B';
	cp=(client_t *)pg->active;
	while (cp!=NULL) {
		if (cp->status=='1') {
			res=sendto(sock,msg,size,0,(struct sockaddr *)&(cp->cliaddr),sizeof(cp->cliaddr));
		} else {
			if (debug) puts("Zero status found");
		}
		printf("Send res:%d to %s\n",res,inet_ntoa(cp->cliaddr.sin_addr));
		cp=cp->next_c;
	}

/*		
	for (i=0;i<CLISIZE;i++) {
	   if (clidb[i].status=='1') {
	     if ((timenow-clidb[i].t) >STALETIME) {
	        clidb[i].status='0';
	        if (debug) printf("Timeout %d %d %d\n",i,timenow,clidb[i].t);
	        }
	     else {
	        res=sendto(sock,msg,size,0,(struct sockaddr *)&clidb[i].cliaddr,
	                                                  sizeof(clidb[i].cliaddr));
	        if (debug) {
	             printf("Send res:%d to %s\n",res,inet_ntoa(clidb[i].cliaddr.sin_addr));
	             }
	        }
	    }
	   } 
*/
}
	
	
	
	
usage()
	{
	printf("winpubd [-p port] [-d] [-i]\n");
}
	



stringify(char *ar,u_long size)
{
	while ((size-->1)&&(isascii(*ar))) ar++;
	*ar='\0';
}
